#!/usr/local/bin/perl
#
#   nph-proxy.cgi-- CGI Proxy 1.0: an HTTP proxy in the form of a CGI script.
#     Retrieves the resource at any URL, updating embedded URLs in HTML
#     files to point back through this script.
#
#   Copyright (C) 1996, 1998 by James Marshall, james@jmarshall.com
#
#   For the latest, see http://www.jmarshall.com/tools/cgiproxy/
#
#   When called with a URL in the extra path info, it will return the
#     resource at that URL, as if the URL had been called directly (like
#     an HTTP proxy).  No user information is passed to the server.
#   If the retrieved resource is HTML, then all URLs referenced in it
#     will be modified to retrieve the URLs through this script again.
#   Form submission is supported, with either GET or POST.
#   The owner of the script can choose to forward only text resources (MIME
#     type text/*).  This may help when bandwidth is limited.  Of course,
#     this can still be circumvented by uuencode, etc., but should greatly
#     cut down traffic while still delivering textual information.
#
#
#   CONFIGURATION:
#     None required.  Options:
#       . To restrict to text-only forwarding, set $textonly=1 below.
#       . If you're using another HTTP proxy, set $ENV{'http_proxy'} and
#         $ENV{'no_proxy'} below.
#       . To make this run in Perl 4, see the instructions above the 
#         "use Socket" line below.
#
#     This script MUST be installed as a non-parsed header (NPH) script.
#
#   TO USE:
#     Either call this script with the target URL in the PATH_INFO, or 
#       use the associated startproxy.cgi.
#
#
#   I first wrote this in 1996 as an experiment to allow anonymous browsing.
#     Parts of the code have been upgraded since then, like the addition of
#     HTML 4.0 tags.  The original seed was a program I wrote for Rich
#     Morin's article in the June 1996 issue of Unix Review, online at
#     http://www.cfcl.com/tin/P/9606.html.
#
#   Confession:  I didn't originally write this with the spec for HTTP proxies
#     in mind, and there are probably some violations of the protocol.  This
#     whole thing doesn't really match the proxy model anyway, so I hereby
#     rationalize that the spec can be widely interpreted here.  If there is
#     demand, I can make it more conformant.
#
#   TO DO:
#     add even the tiniest bit of error checking
#     make able to be a real HTTP proxy, listening to a port (maybe make
#       modified version)
#     support cookies?
#
#

#------ initialization ----------------------------------------------------

# Allow the installer of this script to only forward text files (no images,
#   audio, etc.), for bandwidth concerns.
$textonly= 0 ;      # set to 1 to allow only text data, 0 to allow all

# set HTTP proxy if needed
#$ENV{'http_proxy'}= '' ;
#$ENV{'no_proxy'}= '' ;

# Requires Perl 5.  To run in Perl 4, or for more speed, remove this line and 
#   hard-code $AF_INET and $SOCK_STREAM (usually in /usr/include/sys/socket.h
#   or /usr/include/linux/socket.h) into &newsocketto().
use Socket ;

# OK, let's time this thing
#($sutime,$sstime)= (times)[0,1] ;

# %urlsin was an associative array that listed tags and attributes that 
#   may contain URLs, declared like:
# %urlsin= ('a', 'href',
#           'applet', 'codebase',
#          'fig', 'src|imagemap',
#          'form', 'action|script',
#              .
#              .
#              .
#          ) ;


# The CGI "standard" at NCSA is unclear on whether $ENV{'SCRIPT_NAME'} 
#   should have a leading slash, and different servers implement
#   it differently.  Here we just make it without a leading slash.
$ENV{'SCRIPT_NAME'}=~ s#^/## ;

# Copy often-used environment vars into scalars, for efficiency
$env_accept= $ENV{'HTTP_ACCEPT'} || '*/*' ;     # may be modified later

# QUERY_STRING with question mark prepended
$qs_out=  $ENV{'QUERY_STRING'} ne ''   ?  '?' . $ENV{'QUERY_STRING'}   :  '' ;

# Calculate $thisurl, useful in many places
$portst=  $ENV{'SERVER_PORT'}==80  ?  ''   :  ':' . $ENV{'SERVER_PORT'} ;
$thisurl= join('', 'http://', $ENV{'SERVER_NAME'}, $portst, 
                    '/', $ENV{'SCRIPT_NAME'}, '/') ;


#------ parsing of URL and other input ------------------------------------

# Read the URL from PATH_INFO, stripping leading slash
($URL= $ENV{'PATH_INFO'})=~ s#^/## ;

# Parse the URL, simply.  Note that "#" fragments are not sent by the browser
#   and "?" queries are sent in QUERY_STRING.  The occasional ";" param will
#   be included in $path, and handled correctly.
($scheme, $host, $port, $path)= 
    ($URL=~ m#^([\w+.-]+)://([^/:]*):?([^/]*)(/.*)?$#i) ;
$port || ($port= 80) ;

# Alert the user to non-HTTP URL, with an intermediate page
&nonHTTPwarning($URL.$qs_out) unless ($scheme=~ /^http$/i) ;

# If path is empty, send back Location: to include the final slash.
#   Otherwise, the browser itself will resolve relative URL's wrong.
if ($path eq '') {
    print "HTTP/1.0 302 Found\012Location: ", $thisurl, $URL, "/\012\012" ;
    exit ;
}


# Exclude non-text if it's not allowed.  Err on the side of allowing too much.
if ($textonly) {
    # First, forbid requests for filenames with non-text-type extensions
    # Extensions are roughly taken from Netscape's "Helper Preferences" screen
    #   (but that was in 1996).
    # Boy, I really need that list of binary file extensions.  Write me if
    #   you have one (August 1998).
    $nontext= 'gif|jpeg|jpe|jpg|tiff|tif|png|bmp|xbm'   # images
            . '|mp2|mp3|wav|aif|aiff|au|snd'            # audios
            . '|avi|qt|mov|mpeg|mpg|mpe'                # videos
            . '|gz|Z|exe|gtar|tar|zip|sit|hqx|pdf'      # applications
            . '|ram|rm|ra' ;                            # others

    &nontextdie if ($path=~ /\.($nontext)(;|$)/i) ;

    # Then, filter the "Accept:" header to accept only text
    $env_accept=~ s#\*/\*#text/*#g ;    # not strictly perfect
    $env_accept= join(', ', grep(m#^text/#i, split(/\s*,\s*/, $env_accept)) ) ;
    &nontextdie unless $env_accept ne '' ;
}


$realhost= $host ;
$realport= $port ;
$realpath= $path ;

# there must be a smoother way to handle proxies....
if ($ENV{'http_proxy'}) {
    local($dontproxy) ;
    foreach (split(/\s*,\s*/, $ENV{'no_proxy'})) {
        last if ($dontproxy= $host=~ /$_$/) ;
    }
    unless ($dontproxy) {
	# could be slightly more efficient in Perl 5
        ($dummy,$realhost,$realport)=
            $ENV{'http_proxy'}=~ m#^(http://)?([^/:]*):?([^/]*)#i ;
        $realport= ($realport || 80) ;
        $realpath= $URL ;
    }
}


#------ Connect socket to host; send request; wait with select() ----------

# Print request line, plus Accept: and User-Agent: headers.
# We're anonymously browsing, so don't include From: or Referer: headers.
# Use correct request method, path, and query string if it exists.
&newsocketto(*S, $realhost, $realport) ;

print S $ENV{'REQUEST_METHOD'}, ' ', $realpath, $qs_out, " HTTP/1.0\015\012",
	'Host: ', $host, ':', $port, "\015\012",
        'Accept: ', $env_accept, "\015\012",        # possibly modified above
        'User-Agent: ', $ENV{'HTTP_USER_AGENT'}, "\015\012" ;

# If request method is POST, copy content headers and body to request.  Loop
#   to guarantee all is read from STDIN.
if ($ENV{'REQUEST_METHOD'} eq 'POST') {
    $lefttoget= $ENV{'CONTENT_LENGTH'} ;
    print S 'Content-type: ', $ENV{'CONTENT_TYPE'}, "\015\012", 
            'Content-length: ', $lefttoget, "\015\012\015\012" ;
    do {
        $lefttoget-= read(STDIN, $postbody, $lefttoget) ;
        print S $postbody ;
    } while $lefttoget && length($postbody) ;

# For GET requests, just add extra blank line
} else {
    print S "\015\012" ;
}

vec($rin= '', fileno(S), 1)= 1 ;
select($rin, undef, undef, 60) 
    || &HTMLdie("No response from $realhost:$realport") ;


#------ Read full response into $status, $headers, and $_ (body) ----------

# Full elapsed time, not just CPU time
# $starttime= time ;

# Support both HTTP 1.x and HTTP 0.9
$status= <S> ;  # first line, which is the status line in HTTP 1.x

# HTTP 1.x
if ($status=~ m#^HTTP/#) {
    do {
        $headers.= $_= <S> ;    # $headers includes last blank line
    } until (/^(\015\012|\012)$/) ;   # lines may be terminated with LF or CRLF

    # Unfold long header lines, a la RFC 822 section 3.1.1
    $headers=~ s/(\015\012|\012)[ \t]/ /g ;

    # If we're text only, then cut off non-text responses
    if ($textonly) {
        $*= 1 ;
        if ($headers=~ m#^Content-type:\s*([\w/]*)#i) {
            (close(S), &nontextdie) unless $1=~ m#^text/#i ;
        }
    }

    $/= '>' ;
    @body= <S> ;

# HTTP 0.9 
} else {
    undef $/ ;
    $_= $status . <S> ;
    $status= '' ;

    # split through ">", including "delimiters", and remove (via grep)
    #   the "actual" matches, which are blank
    @body= grep(length,split( /([^>]*>?)/ )) ;
}

close(S) ;


#------ Modify entire response to point back through this script ----------

# NOTE: IT IS IMPORTANT TO DO THIS AS COMPLETELY AS POSSIBLE.  IF A
# USER UNKNOWINGLY GOES TO PAGE DIRECTLY AND NOT THROUGH THIS PROXY,
# HE/SHE MAY REVEAL HIM/HERSELF IN AN UNINTENDED WAY.

#--------------------------------------------------------------------------
# These were notes to myself from testing the speed of different methods.
#   Names like "nph-proxy2" refer to different modifications.  This version,
#   the fastest, was called nph-proxy2b.
#--------------------------------------------------------------------------
#
# [This version is nph-proxy2b:  Break into @body array, do not use 
#   %urlsin to test if tags should be updated, replacement strings 
#   use (a|b|c) syntax.]
#
# This is by far the most time-consuming part of the script, the updating
#   of all URLs in an HTML file.
#
# Results of informal speed testing (not what I expected):
#   Breaking into @body array instead of one big string definitely saves 
#       significant time-- compare nph-proxy2 to nph-proxy1.  One test
#       showed a time saving of 1/3 to 1/2.
#   Using a %urlsin array does NOT seem to save time, even when only used
#       as boolean test to see if tag might contain URL-- compare nph-proxy2
#       to nph-proxy3, nph-proxy4, and (most similar) nph-proxy5.  Oh well.
#   It seems that reading one tag at a time, converting, and sending it
#       through does NOT save elapsed time over reading all tags before
#       converting, like I thought it would.  Both the CPU and elapsed
#       time are longer for one-tag-at-a-time approach-- compare
#       nph-proxy2 and nph-piper2.
#
#   Results of nph-proxy2 (blocks for each tag) to nph-proxy2b 
#   (single "(att1|att2|att3)" style regex) testing:
#       Mixed results, but overall, using a single regex (nph-proxy2b)
#       takes less CPU "user time", about the same "system time", and
#       slightly more elapsed time than nph-proxy2; I don't know why.
#       The elapsed time is slightly more both within the script and at
#       the HTTP client's end.  All differences are less than 10% on
#       average, and nph-proxy2b occasionally shows LESS elapsed time
#       than nph-proxy2.  CPU time is always less for nph-proxy2b than
#       for nph-proxy2.
#   Since the bottleneck of CPU time is tighter than for elapsed time,
#       and since the elapsed-time loss is less than the CPU time gain,
#       let's go with nph-proxy2b (not that it really makes much difference).
#       Besides, the code is easier to read that way.
#
# So basically, breaking into @body array helps, but not much else does.
#
#--------------------------------------------------------------------------
# 8-4-98 JSM: Found a bug in the regex, so changed it.  It more resembles
#   nph-proxy2.cgi now.
#--------------------------------------------------------------------------

$*= 1 ;     # allow multi-line matching

# Set $basehost correctly-- first see if there's a <base> tag, then if
#   there's a Location: header; otherwise, use original URL.
# This is part of &fullurl(), placed here for speed.
($_)= grep(/<\s*base\b/i, @body) ;
if      ( ($basehost) = m#<\s*base\b[^>]*\bhref\s*=\s*"?([\w+.-]+://[^/\s">]+)#i ) {
} elsif ( ($basehost)= ($headers=~ m#^Location:\s*([\w+.-]+://[^/\s]+)#i) ) {
} else  { ($basehost= join('', 'http://', $host, (($port==80) ?'' :":$port") ) ) }

$basehost= $thisurl . $basehost ;


# If we get a 300-level response code, update the Location: header to point
#   back through the script, so the browser will retrieve it correctly.
if ($status=~ m#^HTTP/[0-9.]*\s*3\d\d#) {
    $headers=~ s/^Location:\s*(.*)/'Location: ' . &fullurl($1)/gie ;
    $headers=~ s/^URI:\s*(.*)/     'URI: '      . &fullurl($1)/gie ;
}



# Update all URLs in all tags that refer to URLs
# Only update the URLs if it's HTML (or using HTTP 0.9), and if it's not
#   empty.

if ( (($headers=~ m#^Content-type:\s*text/html#i) || !$headers) 
     && ($body[0] ne '') ) {

    # This is a complete list of HTML tags/attributes that may include a URL,
    #   compiled from the list at Sandia (read on 9-17-96, but last updated on
    #   12-7-95), and the latest HTML DTD as of 9-17-96.  The Sandia list
    #   includes non-standard tags found to be used by Netscape or Microsoft.
    # List updated 7-31-98 by JSM to include all URL-type attributes defined
    #   in HTML 4.0.
    # If anyone knows of a well-maintained list of standard and non-standard
    #   tags with URLs in them, please let me know!!
    # Note that most of these are very rarely used, if ever.  They're included
    #   for safety, since we don't want an anonymous user accidentally
    #   revealing themselves because of a non-anonymized URL.
    # This regex favors speed over readability.
    # This could probably be more efficient, but the solution eludes me.
    # Note that this is not a perfect regex.  The substitutions can be fooled
    #   by, e.g., <input value="<a href=bad"> or <input value="src=bad">.
    # The regex also terminates URLs at spaces; to fix this would take a
    #   much slower and messier regex.
    # Denoting which of these are for images/binaries might be helpful, if
    #   we need more elaborate text-only support.
    # 8-4-98 JSM:  Changed the regex to fix a bug: would only change the 
    #   first URL in a tag.  New regex is probably slower.

    foreach (@body) {

	# Put the most common cases first

        s/(<[^>]*\bhref\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*a\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\blowsrc\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\blongdesc\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
        s/(<[^>]*\busemap\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\bdynsrc\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
            next if /<\s*img\b/i ;

        s/(<[^>]*\bbackground\s*=\s*"?)([^\s">]*)/ $1 . &fullurl($2) /ie,
            next if /<\s*body\b/i ;

        s/(<[^>]*\bhref\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*base\b/i ;     # has special significance

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\blongdesc\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
            next if /<\s*frame\b/i ;

        s/(<[^>]*\baction\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\bscript\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
            next if /<\s*form\b/i ;     # needs special attention

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\busemap\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
            next if /<\s*input\b/i ;

        s/(<[^>]*\bhref\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*area\b/i ;

        s/(<[^>]*\bcodebase\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
        s/(<[^>]*\bcode\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
        s/(<[^>]*\bobject\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\barchive\s*=\s*"?)([^\s">]*)/    $1 . &fullurl($2) /ie,
            next if /<\s*applet\b/i ;


        # These are seldom-used tags, or tags that seldom have URLs in them

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*bgsound\b/i ;  # Microsoft only

        s/(<[^>]*\bcite\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*blockquote\b/i ;

        s/(<[^>]*\bcite\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*del\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*embed\b/i ;    # Netscape only

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\bimagemap\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
            next if /<\s*fig\b/i ;      # HTML 3.0

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*h[1-6]\b/i ;   # HTML 3.0

        s/(<[^>]*\bprofile\s*=\s*"?)([^\s">]*)/    $1 . &fullurl($2) /ie,
            next if /<\s*head\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*hr\b/i ;       # HTML 3.0

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\blongdesc\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
            next if /<\s*iframe\b/i ;

        s/(<[^>]*\bcite\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*ins\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*layer\b/i ;

        s/(<[^>]*\bhref\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
        s/(<[^>]*\burn\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*link\b/i ;

        s/(<[^>]*\burl\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*meta\b/i ;     # Netscape only

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*note\b/i ;     # HTML 3.0

        s/(<[^>]*\busemap\s*=\s*"?)([^\s">]*)/     $1 . &fullurl($2) /ie,
        s/(<[^>]*\bcodebase\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
        s/(<[^>]*\bdata\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
        s/(<[^>]*\barchive\s*=\s*"?)([^\s">]*)/    $1 . &fullurl($2) /ie,
        s/(<[^>]*\bclassid\s*=\s*"?)([^\s">]*)/    $1 . &fullurl($2) /ie,
        s/(<[^>]*\bname\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*object\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\bimagemap\s*=\s*"?)([^\s">]*)/   $1 . &fullurl($2) /ie,
            next if /<\s*overlay\b/i ;  # HTML 3.0

        s/(<[^>]*\bcite\s*=\s*"?)([^\s">]*)/       $1 . &fullurl($2) /ie,
            next if /<\s*q\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
        s/(<[^>]*\bfor\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*script\b/i ;

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*select\b/i ;   # HTML 3.0

        s/(<[^>]*\bsrc\s*=\s*"?)([^\s">]*)/        $1 . &fullurl($2) /ie,
            next if /<\s*ul\b/i ;       # HTML 3.0

    }   # foreach (@body)


    $body[0]= "<!-- resource has been modified by proxy script -->\n" 
            . $body[0] ;

    # Change Content-Length header, since we're changing the content
    $headers=~ s/^Content-Length:.*\012/ 'Content-Length: ' 
            . (grep($newlength+=length(),@body), $newlength) 
            . "\015\012"/ie ;

}

# print the status line, headers, and the entire (modified) resource
print $status, $headers, @body ;


# Put this back in to run speed trials
#if (($headers=~ m#^Content-type:\s*text/html#i) || !$headers) {
#    # OK, let's time this thing
#    ($eutime,$estime)= (times)[0,1] ;
#    open(LOG,">>proxy2b.log") ;
#    print LOG "full times: ", $eutime-$sutime, " ", $estime-$sstime, 
#        " ", time-$starttime, "  URL: $URL\n" ;
#    close(LOG) ;
#}


exit ;


# Returns the full URL to query our script for an absolute URL.  Includes an
#   abbreviation of relative->absolute URL conversion.  It handles almost all
#   URLs that actually exist, though it does not contract "." and ".." paths.
# URLs it does not handle include "//www.host.com", "http:localpath/file",
#   and other incomplete (but valid) relative URLs.  Can be fixed if needed,
#   but is kept like this for efficiency.
# The full procedure is described in RFC 1808, section 4.
# Note that the calculation of $basehost above is an integral part of this 
#   routine, placed above for speed.
# Note that because of the structure of extra path information in the URL,
#   valid relative URLs will be resolved correctly by the browser.  :):)
# BUG: If a (non-HTTP) URL has a query string with an actual space in it
#   (like with "mailto:...?subject=a b"), the query string is truncated when
#   the user follows the converted link, because the browser will terminate 
#   the URL at the space.  This is messy and costly to fix, unfortunately--
#   would require replacing spaces with %20 in the query string part of
#   $relurl here; also would require messier regex in tag/attribute checking
#   above.
sub fullurl {
    local($relurl)= @_ ;
    $relurl=~ m#^[\w+.-]*:#i  && return ($thisurl.$relurl) ;  # absolute URL
    $relurl=~ m#^/#           && return ($basehost.$relurl) ; # absolute path, relative URL
                                 return $relurl ;             # relative URL
}


#------ Utility and output routines ---------------------------------------

# The following subroutine looks messy, but can be used to open any
#   TCP/IP socket in any Perl program.  Except for the &HTMLdie() part.
sub newsocketto {
    local(*S, $host, $port)= @_ ;

    # Create the remote host data structure, from host name or IP address
    $hostaddr= ($host=~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/)
                  ?  pack('c4', $1, $2, $3, $4)     # for IP address
                  :  ( (gethostbyname($host))[4]    # for alpha host name
                       || &HTMLdie("Couldn't find address for $host: $!") ) ;
    $remotehost= pack('S n a4 x8', AF_INET, $port, $hostaddr) ;

    # Create the socket and connect to the remote host
    socket(S, AF_INET, SOCK_STREAM, (getprotobyname('tcp'))[2])
        || &HTMLdie("Couldn't create socket: $!") ;
    connect(S, $remotehost) 
        || &HTMLdie("Couldn't connect to $host:$port: $!") ;
    select((select(S), $|=1)[0]) ;      # unbuffer the socket
}


# Alert the user to non-HTTP URL, with this intermediate page
sub nonHTTPwarning {
    print <<EOF ;
HTTP/1.0 200 OK
Content-type: text/html

<html>
<head><title>WARNING: Entering non-anonymous area!</title></head>
<body>
<h1>WARNING: Entering non-anonymous area!</h1>
<h3>This proxy only supports HTTP.  Any browsing to a non-HTTP URL will
be directly from your browser, and no longer anonymous.</h3>
<h3>Click the link below to continue to the URL, non-anonymously.</h3>
<blockquote><tt><a href="$_[0]">$_[0]</a></tt></blockquote>
<p>
<hr>
<a href="http://www.jmarshall.com/tools/cgiproxy/"><i>CGIProxy 1.0</i></a>
<p>
</body>
</html>
EOF

    exit ;
}


# Return "403 Forbidden" message, with explanatory text
sub nontextdie {
    print <<EOF ;
HTTP/1.0 403 Forbidden
Content-type: text/html

<html>
<head><title>Proxy cannot forward non-text files</title></head>
<body>
<h1>Proxy cannot forward non-text files</h1>
<p>Due to bandwidth limitations, the owner of this particular proxy is
forwarding only text files.  For best results, turn off automatic image 
loading if your browser lets you.
<p>If you need access to images or other binary data, route your browser 
through another proxy (or install one yourself--
<a href="http://www.jmarshall.com/tools/cgiproxy/">it's easy</a>).
<p>
<hr>
<a href="http://www.jmarshall.com/tools/cgiproxy/"><i>CGIProxy 1.0</i></a>
<p>
</body>
</html>
EOF

    exit ;
}


# Die, outputting HTML error page
sub HTMLdie {
    local($msg)= @_ ;
    print <<EOF ;
HTTP/1.0 200 OK
Content-Type: text/html

<html>
<head><title>CGI Proxy Error</title></head>
<body>
<h1>CGI Proxy Error</h1>
<h3>$msg</h3>
<p>
<hr>
<a href="http://www.jmarshall.com/tools/cgiproxy/"><i>CGIProxy 1.0</i></a>
<p>
</body>
</html>
EOF

    exit ;
}
